# -*- coding: UTF-8 -*-
""""
Created on 04.12.21
Contains parallel functor map.

:author:     Martin Dočekal
"""
import multiprocessing
import queue
from multiprocessing import Process, Queue
from typing import TypeVar, Iterable, Generator, Callable

from windpyutils.buffers import Buffer

T = TypeVar('T')
R = TypeVar('R')


class FunctorWorker(Process):
    """
    Functor worker for FunctorMap.
    """

    def __init__(self, pf: Callable[[T], R], work_queue: Queue, results_queue: Queue):
        """
        Initialization of parallel worker.

        :param pf: Function you want to run in data-parallel way.
        :param work_queue: queue that is used for receiving work and stop orders
        :param results_queue: queue that is used for sending results
        """
        super().__init__()
        self.pf = pf
        self._work_queue = work_queue
        self._results_queue = results_queue

    def run(self) -> None:
        """
        Run the process.
        """
        try:
            while True:
                q_item = self._work_queue.get()

                if q_item is None:
                    # all done
                    break

                i, data_list = q_item

                self._results_queue.put((i, [self.pf(x) for x in data_list]))

        finally:
            self._work_queue.close()
            self._results_queue.close()


class FunctorMap:
    """
    A parallel map that uses given function.
    """

    def __init__(self, pf: Callable[[T], R], workers: int = -1):
        """
        Initialization of parallel functor map.

        :param pf: Function you want to run in data-parallel way.
        :param workers: Number of parallel workers.
            Values <=0 will create number of workers that will be same as number of cpus.
        """
        super().__init__()
        if workers <= 0:
            workers = multiprocessing.cpu_count()

        self._work_queue = Queue(workers)
        self._results_queue = Queue()
        self.procs = [
            FunctorWorker(pf=pf, work_queue=self._work_queue, results_queue=self._results_queue) for _ in range(workers)
        ]

    def __enter__(self) -> "FunctorMap":
        for p in self.procs:
            p.daemon = True
            p.start()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        for _ in range(len(self.procs)):
            self._work_queue.put(None)
        for p in self.procs:
            p.join()

    def __call__(self, data: Iterable[T], chunk_size: int = 1) -> Generator[R, None, None]:
        """
        Applies functor on each element in iterable.
        honors the order

        :param data: iterable of data that should be passed to functor
        :param chunk_size: size of a chunk that is send to a process
        :return: generator of results
        """

        buffer = Buffer()

        def chunking(d):
            ch = []
            for x in d:
                ch.append(x)
                if len(ch) == chunk_size:
                    yield ch
                    ch = []
            if len(ch) > 0:
                yield ch

        data_cnt = 0
        finished_cnt = 0
        for i, chunk in enumerate(chunking(data)):
            self._work_queue.put((i, chunk))
            data_cnt += 1

            try:
                # read the results
                while True:
                    res_i, res_chunk = self._results_queue.get(False)
                    for ch in buffer(res_i, res_chunk):
                        finished_cnt += 1
                        for x in ch:
                            yield x

            except queue.Empty:
                pass

        while finished_cnt < data_cnt:
            res_i, res_chunk = self._results_queue.get()
            for ch in buffer(res_i, res_chunk):
                finished_cnt += 1
                for x in ch:
                    yield x



# -*- coding: UTF-8 -*-
""""
Created on 27.01.21

Module containing buffers.

:author:     Martin Dočekal
"""
from typing import TextIO, Generator, Any


class Buffer:
    """
    Buffer for remaining order.
    For each data item expects its position (starting from 0).

    Example:
        >>> b = Buffer()
        >>> b(1, "B")
        >>> b(0, "A")
        >>> b(2, "C")
        >>> b(4, "E")
        >>> b.waiting_for()
        0
        >>> list(b)
        ["A", "B", "C"]
        >>> b.waiting_for()
        3
        >>> b(3, "D")
        >>> list(b)
        ["D", "E"]
    """

    def __init__(self):
        self._storage = {}
        self._waiting_for = 0

    def waiting_for(self) -> int:
        """
        Position of item that the buffer is waiting for to generate next.
        """
        return self._waiting_for

    def __call__(self, i: int, x: Any) -> "Buffer":
        """
        Adds new item into buffer.
        If you place an item with same i, before buffering, the new item wil overwrite the old one.

        :param i: number for determining order
        :param x: data
        :return: returns itself
        :raise AttributeError: when you place already generated position
        """

        if i < self._waiting_for:
            raise AttributeError("Already generated position is placed.")

        self._storage[i] = x
        return self

    def __iter__(self) -> Generator[Any, None, None]:
        """
        Generates subsequence that is in order already

        :return: generator of buffered items
        """
        while self._waiting_for in self._storage:
            yield self._storage[self._waiting_for]
            del self._storage[self._waiting_for]
            self._waiting_for += 1

    def flush(self):
        """
        Flushes the buffer.
        """
        self._storage = {}
        self._waiting_for = 0


class PrintBuffer:
    """
    Stores and prints data marked with serial numbers in sorted order.

    Use unique serial numbers else the data will be overwritten.

    Usage:

        pBuffer = PrintBuffer()
        pBuffer.print(1, "B")   # stores B because the next printed string should have serial number 0
        pBuffer.print(2, "C")   # stores C because the next printed string should have serial number 0
        pBuffer.print(0, "A")   # prints
        ->  A
            B
            C
        pBuffer.print(3, "C")  # print immediately because 3 is the next serial number
        ->  C

    :var file_out: Where the output should be printed.
    :var print_flush: flush parameter for print method
    :var end: String appended to the last character before print. Default is a newline.
    """

    def __init__(self, file_out: TextIO, print_flush: bool = False, end: str = "\n"):
        """
        initialization of buffer

        :param file_out: Where the output should be printed.
        :type file_out: TextIO
        :param print_flush: flush parameter for print method
        :type print_flush: bool
        :param end: String appended to the last character before print. Default is a newline.
        :type end: str
        """
        self._waiting_for = 0
        self._buffer = {}
        self.fileOut = file_out
        self.printFlush = print_flush
        self.end = end

    def _print(self, value: str):
        """
        Just wrapper for print.

        :param value: Value that should be printed.
        :type value: str
        """
        print(value, file=self.fileOut, flush=self.printFlush, end=self.end)

    @property
    def waiting_for(self) -> int:
        """
        Serial number the buffer is waiting for.
        """
        return self._waiting_for

    def print(self, serial_number: int, value: str) -> bool:
        """
        Prints value if serial number is the one it is waiting for, else stores that value for later.
        If serial number is the one it is waiting for and it also has stored values with consequent serial numbers
        it will print them too.

        Use unique serial numbers else the data will be overwritten.

        :param serial_number: Serial number of actual value.
        :type serial_number: int
        :param value: String that can be printed.
        :type value: str
        :return: True if value (or multiple values) was printed. False when the value was buffered.
        :rtype: bool
        """

        if serial_number == self._waiting_for:
            self._print(value)

            self._waiting_for += 1

            while self._waiting_for in self._buffer:
                self._print(self._buffer[self._waiting_for])
                del self._buffer[self._waiting_for]
                self._waiting_for += 1

            return True
        else:
            self._buffer[serial_number] = value
            return False

    def flush(self):
        """
        Prints out all data stored in buffer in ascending sorted order according to their serial number.

        Sets the waitingFor to the next serial number after the biggest one in memory.
        Do nothing if buffer is empty.
        """
        serial_number = self._waiting_for - 1  # will guarantee that if the buffer is empty the waitingFor does not change

        for serial_number in sorted(self._buffer.keys()):
            self._print(self._buffer[serial_number])
            del self._buffer[serial_number]

        self._waiting_for = serial_number + 1

    def __len__(self):
        """
        Number of stored items.
        """

        return len(self._buffer)

    def clear(self):
        """
        Clears buffer without printing items inside and set waitingFor to zero.
        """
        self._buffer = {}
        self._waiting_for = 0



# -*- coding: UTF-8 -*-
""""
Created on 20.11.19
This module contains implementation of observable singleton logger.

:author:     Martin Dočekal
"""
from .design_patterns import Observable, Singleton


class Logger(Observable, metaclass=Singleton):
    """
    This singleton class is useful for broadcasting logs to observes that registers to
    log method.

    If you want to log something just call Logger().log("something") and all
    observers, registered with Logger().register_observer("LOG", observer_callback_method) method,
    will be called.

    """

    @Observable._event("LOG", True)
    def log(self, txt: str):
        """
        Make a log.

        :param txt: Text of the log.
        :type txt: str
        """
        pass



# -*- coding: UTF-8 -*-
""""
Created on 30.06.20
Module with wrapper that makes arguments parsers raising exceptions.

:author:     Martin Dočekal
"""

from argparse import ArgumentParser


class ArgumentParserError(Exception):
    """
    Exceptions for argument parsing.
    """
    pass


class ExceptionsArgumentParser(ArgumentParser):
    """
    Argument parser that uses exceptions for error handling.
    """

    def error(self, message):
        raise ArgumentParserError(message)




# -*- coding: UTF-8 -*-
""""
Created on 20.11.19
This module contains implementaton of some common design patterns.

:author:     Martin Dočekal
"""
from functools import wraps
from typing import Dict, Callable


class Singleton(type):
    """
    Metaclass for singletons.
    """
    _clsInstances = {}
    """Dict containing instances of all singleton classes."""

    def __call__(cls, *args, **kwargs):
        try:
            return cls._clsInstances[cls]
        except KeyError:
            cls._clsInstances[cls] = super().__call__(*args, **kwargs)
            return cls._clsInstances[cls]


class Observable(object):
    """
    Implementation of observer like design pattern.

    Example Usage:

        class A(Observable):
            def __init__(self):
                super().__init__()

            @Observable.event("STARTS")
            def starts_the_engine(self):
                ...

            @Observable.event("END", True)    #true means that all arguments will be passed to observer
            def end_the_engine(self, data):
                ...

        a=A()
        a.register_observer("STARTS", observer_callback_method)
    """

    @staticmethod
    def event(tag, pass_arguments=False):
        """
        Use this decorator to mark methods that could be observed.
        """
        return Observable._event(tag, pass_arguments)

    @staticmethod
    def _event(tag, pass_arguments=False):
        """
        Use this decorator to mark methods that could be observed.
        """

        def tags_decorator(f):
            @wraps(f)
            def func_wrapper(o, *arg, **karg):
                f(o, *arg, **karg)
                if pass_arguments:
                    o._Observable__notify(tag, *arg, **karg)
                else:
                    o._Observable__notify(tag)

            return func_wrapper

        return tags_decorator

    def __init__(self):
        self.__observers = {}

    @property
    def observers(self):
        """
        Get all observers.
        """

        return self.__observers

    @observers.setter
    def observers(self, observers: Dict[str, Callable]):
        """
        Set new observers.

        :param observers: New observers.
        :type observers:Dict[str,Callable]
        """

        self.__observers = observers

    def clear_observers(self):
        """
        Clears all observers.
        """
        self.__observers = {}

    def register_observer(self, event_tag, observer):
        """
        Register new observer for observable method (_event).

        :param event_tag: The tag that is passed as parameter for _event decorator.
        :type event_tag: str
        :param observer: Method that should be called
        :type observer: Callable
        """

        s = self.__observers.setdefault(event_tag, set())
        s.add(observer)

    def unregister_observer(self, event_tag, observer):
        """
        Unregister observer for observable method (_event).

        :param event_tag: The tag that is passed as parameter for _event decorator.
        :type event_tag: str
        :param observer: Method that should no longer be called
        :type observer: Callable
        """

        try:
            self.__observers[event_tag].remove(observer)
        except KeyError:
            pass

    def __notify(self, event_tag, *arg, **kw):
        """
        Notify all obervers for given method.

        :param event_tag: The tag that is passed as parameter for _event decorator.
        :type event_tag: str
        """
        try:
            for o in self.__observers[event_tag]:
                o(*arg, **kw)
        except KeyError:
            pass



# -*- coding: UTF-8 -*-
""""
Created on 22.03.21
Utils for work with files.

:author:     Martin Dočekal
"""


class RandomLineAccessFile:
    """
    Allows fast access to any line in given file.
    This structure is just for reading.

    Makes line offsets index in advance.

    Example:

        with RandomLineAccessFile("example.txt") as lines:
            print(lines[150])
            print(lines[0])

    :ivar path_to: path to file
    :vartype path_to: str
    :ivar file: file descriptor
    :vartype file: Optional[TextIO]
    """

    def __init__(self, path_to: str):
        """
        initialization
        Makes just the line offsets index. Whole file itself is not loaded into memory.

        :param path_to: path to file
        """

        self.path_to = path_to
        self.file = None
        self._line_offsets = None
        self._index_file()

    def _index_file(self):
        """
        Makes index of line offsets.
        """

        self._line_offsets = [0]

        with open(self.path_to, "rb") as f:
            while f.readline():
                self._line_offsets.append(f.tell())

        del self._line_offsets[-1]

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __len__(self) -> int:
        """
        Number of lines in the file.

        :return: Number of lines in the file.
        """
        return len(self._line_offsets)

    def open(self) -> "RandomLineAccessFile":
        """
        Open the file if it was closed, else it is just empty operation.

        :return: Returns the object itself.
        :rtype: RandomLineAccessFile
        """

        if self.file is None:
            self.file = open(self.path_to, "r")
        return self

    def close(self):
        """
        Closes the file.
        """

        if self.file is not None:
            self.file.close()
            self.file = None

    def __getitem__(self, n) -> str:
        """
        Get n-th line from file.

        :param n: line index
        :return: n-th line
        :raise RuntimeError: When the file is not opened.
        """
        if self.file is None:
            raise RuntimeError("Firstly open the file.")

        self.file.seek(self._line_offsets[n])
        return self.file.readline()



# -*- coding: UTF-8 -*-
""""
Created on 15.10.20
Module containing attribute drive dictionary.

:author:     Martin Dočekal
"""
from keyword import iskeyword


class AttributeDrivenDictionary(dict):
    """
    Base class for all classes that want's to act like a dictionary, but it allows to  access keys like instance variable.

    WARNING: interface may not be compatible with standard dict interface in all cases

    Example of usage:

        class MyDataClass(AttributeDrivenDictionary):
            def __init__(a:str)"
                self.a = a

        m = MyDataClass("hello")
        m.a     # hello
        m["a"]  # hello
    """

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        super().__setitem__(key, value)

    def __setitem__(self, key, value):
        if not isinstance(key, str) or not key.isidentifier() or iskeyword(key) or key == "None":
            # key is in invalid form
            raise KeyError(key)

        super().__setitem__(key, value)

    @property
    def __dict__(self):
        # this is dict so it is mapping object to itself all of it's writable attributes
        return self



# -*- coding: UTF-8 -*-
""""
Created on 13.05.20
This module contains structures that are useful for mocking.

:author:     Martin Dočekal
"""
from abc import ABC, abstractmethod
from itertools import count
from typing import Sequence, Union, Iterator

from windpyutils.generic import RoundSequence


class BasicMockedRand(ABC):
    """
    Basic abstract class for mock rands.
    Every class must have gen parameter returning iterator that on next call return next value.
    """

    def __call__(self, *args, **kwargs):
        return next(self.gen)

    @property
    @abstractmethod
    def gen(self) -> Iterator:
        pass

    def sample(self):
        """
        Same as a simple call.
        """
        return self()


class MockedRand(BasicMockedRand):
    """
    Mock that has deterministic rand like behaviour.

    It can operate in two modes:
        Steps
            Init example: m = MockedRand(0.1)
                0   0.1     0.2 ...
            Every call returns value:
                N*STEP - int(N*STEP)
            Where N is zero, first to n-th call. STEP is parameter provided in the initialization.

        Sequence
            Init example: m = MockedRand([0.1, 0.2, 0.6])
            Every call returns next value from provided collection. When reaches the end it start from beginning again.
    """

    class Counter(object):

        def __init__(self, step):
            self.step = step
            self.cnt = -1

        def __next__(self):
            self.cnt += 1
            tmp = self.step * self.cnt
            return tmp - int(tmp)

    def __init__(self, use: Union[Sequence[float], float] = None):
        """
        Initialization of he mock.

        :param use: Collection that will be used for generating or generator step.
        :type use: Union[Sequence[float], float]
        """

        if isinstance(use, float) or isinstance(use, int):
            self._gen = self.Counter(use)
        else:
            self._gen = RoundSequence(use)

    @property
    def gen(self):
        return self._gen


class MockedRandInt(BasicMockedRand):
    """
    Mock that has deterministic randint like behaviour.

    It can operate in two modes:
        Default
            Init example: m = MockRandInt()
            Every call returns value of a call counter from 0 (first call returns 0) to N*STEP.

        Sequence
            Init example: m = MockRandInt([1, 1, 0, 0, 2, 0, 1])
            Every call returns next value from provided collection. When reaches the end it start from beginning again.
    """

    def __init__(self, use: Union[Sequence[int], int]):
        """
        Initialization of he mock.

        :param use: Collection that will be used for generating or generator step.
        :type use: Union[Sequence[int], int]
        """

        if isinstance(use, float) or isinstance(use, int):
            self._gen = count(0, use)
        else:
            self._gen = RoundSequence(use)

    @property
    def gen(self):
        return self._gen



# -*- coding: UTF-8 -*-
""""
Created on 08.04.20
Immutable set for spans.

:author:     Martin Dočekal
"""
import copy
import itertools
from abc import ABC, abstractmethod

from typing import Any, Set, Iterable, Tuple, Iterator, Union, Sequence


class SpanSetEqRelation(ABC):
    """
    Abstract class for equality relation that is used in __contains__ method of SpanSet, to determine if span is
    in given set.
    """

    @abstractmethod
    def __call__(self, x_start: Any, x_end: Any, y_start: Any, y_end: Any) -> bool:
        """
        Determines if span x and span y are equal.
        Span x is span on the left side of the in operator and span y is span from investigated set.

        :param x_start: Start of span x.
        :type x_start: Any
        :param x_end: End of span x.
        :type x_end: Any
        :param y_start: Start of span y.
        :type y_start: Any
        :param y_end: End of span y.
        :type y_end: Any
        :return: Returns bool that determines whether x and y spans are equal (true means equal).
        :rtype: bool
        """
        pass


class SpanSetExactEqRelation(SpanSetEqRelation):
    """
    Span equality relation that considers two spans equal only when they are exactly the same.
    """

    def __call__(self, x_start: float, x_end: float, y_start: float, y_end: float) -> bool:
        return x_start == y_start and x_end == y_end


class SpanSetPartOfEqRelation(SpanSetEqRelation):
    """
    Span equality relation that considers two spans equal only when x span is part of y span (x is completely inside y).
    """

    def __call__(self, x_start: float, x_end: float, y_start: float, y_end: float) -> bool:
        return y_start <= x_start and x_end <= y_end


class SpanSetIncludesEqRelation(SpanSetEqRelation):
    """
    Span equality relation that considers two spans equal only when x span includes whole y span.
    """

    def __call__(self, x_start: float, x_end: float, y_start: float, y_end: float) -> bool:
        return x_start <= y_start and y_end <= x_end


class SpanSetOverlapsEqRelation(SpanSetEqRelation):
    """
    Span equality relation that considers two spans equal only when x span y span shares non empty interval
    (x and y overlaps).

    """

    def __call__(self, x_start: float, x_end: float, y_start: float, y_end: float) -> bool:
        # Explanation of condition:
        #   It is easier to start with contradiction, because original relation has much more cases, so...
        #   This relation returns false when there is no overlap:
        #       xs ----------- xe
        #                              ys ----------- ye
        #       xe < ys
        #      or:
        #                              xs ----------- xe
        #        ys ----------- ye
        #       ye < xs
        #
        #   But we are interested in negation, so:
        #   not(xe < ys or ye < xs) =
        #       (xe >= ys) and (ye >= xs)
        return x_end >= y_start and y_end >= x_start


class SpanSet(Set):
    """
    Immutable set for spans.

    Implementation comment to standard set operators:
        Unfortunately standard implementation does not handle the eqRelation and seems to return results for regular
        equality (exact match in terms of SpanSetEqRelation). I think it's probably because it uses __eq__ on elements
        itself and so it does not strictly implements the definitions of this operations, for overloaded ∈. Let me
        explain it on the example.
        Union of two sets A and B A ∪ B is defined as:
            A ∪ B = { x∣ x∈A ∨ x∈B }        (where ∈ are custom operators for set A respective B)
        so there is a problem when ∈ (in operator in python notation that is using __contains__ method) has different
        then ordinary behaviour and as I already said it seems the default implementation does not use the operator on
        set level and probably uses __eq__ method on element level, but this is just my guess that led me to own
        implementation.

    :ivar starts: Spans starts.
    :vartype starts: Sequence
    :ivar ends: Spans ends.
    :vartype ends: Sequence
    :ivar eq_relation: New equal relation that determines equality that is used in __contains__ method
        (and at initialization for similar reason)to determine if span is in given set or more precisely
        that span has a span in this set that is equal to it according to that given relation.
    :vartype eqRelation: SpanSetEqRelation
    """

    def __init__(self, starts: Union[Sequence, Iterable[Tuple[Any, Any]]], ends: Sequence = None,
                 force_no_dup_check: bool = False, eq_relation: SpanSetEqRelation = SpanSetExactEqRelation()):
        """
        Set initialization from iterable of spans or two Sequences. Both Sequence must have the same size,
        because span is composed from offsets on index x in way that span's start is starts[x] and span's end is
        ends[x].

        WARNING: Input Sequences may or not may be reused (not cloned). For more information see forceNoDupCheck param
        documentation.

        :param starts: Spans starts.
            Or iterable of tuples that will be used for initializations and ends will be omitted.
        :type starts: Union[Sequence, Iterable[Tuple[Any, Any]]]
        :param ends: Spans ends.
        :type ends: Sequence
        :param force_no_dup_check: If true than forces to not use duplicate check of spans in inputs and reuses
            input Sequences, which means that if you now that there are no duplicates, than you will get more memory
            efficient set and also the initialization process will be faster.

            This parameter is not obeyed when starts contains Iterable of spans.
        :type force_no_dup_check: bool
        :param eq_relation: New equal relation that determines equality that is used in __contains__ method
        (and at initialization for similar reason)to determine if span is in given set or more precisely
        that span has a span in this set that is equal to it according to that given relation.
        :type eq_relation: SpanSetEqRelation
        """

        super().__init__()
        self.eq_relation = eq_relation

        if ends is not None:
            assert len(starts) == len(ends)
            if force_no_dup_check:
                self.starts = starts
                self.ends = ends
            else:
                # ok we need to look for duplicates
                self.starts = []
                self.ends = []

                for i, s in enumerate(starts):
                    # check if actual span is already in and if not save it to our sequence
                    not_in = True
                    for investigatedSpanIndex in range(len(self.starts)):
                        if self.eq_relation(s, ends[i], self.starts[investigatedSpanIndex],
                                            self.ends[investigatedSpanIndex]):
                            not_in = False
                            break

                    if not_in:
                        self.starts.append(s)
                        self.ends.append(ends[i])

        else:
            self.starts = []
            self.ends = []

            for s, e in starts:
                not_in = True
                for checkIn in range(len(self.starts)):
                    if self.eq_relation(s, e, self.starts[checkIn], self.ends[checkIn]):
                        not_in = False
                        break
                if not_in:
                    self.starts.append(s)
                    self.ends.append(e)

    def __len__(self) -> int:
        return len(self.starts)

    def __str__(self):
        return "{" + ", ".join("({}, {})".format(s, e) for s, e in self) + "}"

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        for s, e in zip(self.starts, self.ends):
            yield s, e

    def __contains__(self, span: Tuple[Any, Any]) -> bool:
        for i, s in enumerate(self.starts):
            if self.eq_relation(span[0], span[1], s, self.ends[i]):
                # span that is equal to this one is in this set
                return True
        return False

    def __le__(self, other: 'SpanSet') -> bool:
        """
        Check if actual set is subset of the other set.

        Definition:
           A ⊆ B <=> ∀x (x∈A => x∈B)

        :param other: The other set.
        :type other: 'SpanSet'
        :return: True when this set is subset of the other set.
        :rtype: bool
        """
        return all(x in other for x in self)

    def __lt__(self, other: 'SpanSet') -> bool:
        """
        Check if actual set is proper subset of the other set.

        Definition:
           A ⊂ B <=> ∀x (x∈A => x∈B) ∧ A  ≠ B

        :param other: The other set.
        :type other: 'SpanSet'
        :return: True when this set is proper subset of the other set.
        :rtype: bool
        """
        return self <= other and self != other

    def __eq__(self, other: 'SpanSet') -> bool:
        """
        Check if actual set is equal with other set.

        Definition:
           A = B <=> A ⊆ B ∧ B ⊆ A

        :param other: The other set.
        :type other: 'SpanSet'
        :return: True when this set equal with the other set.
        :rtype: bool
        """
        return self <= other <= self

    def __ne__(self, other: 'SpanSet') -> bool:
        """
        Check if actual set is not equal with the other set.

        :param other: The other set.
        :type other: 'SpanSet'
        :return: True when this set is not equal with the other set.
        :rtype: bool
        """
        return not (self == other)

    def __ge__(self, other: 'SpanSet') -> bool:
        """
        Check if actual set is super set of the other set.

        :param other: The other set.
        :type other: 'SpanSet'
        :return: True when this set is supper set of the other set.
        :rtype: bool
        """
        return other <= self

    def __gt__(self, other: 'SpanSet') -> bool:
        """
        Check if actual set is proper super set of the other set.

        :param other: The other set.
        :type other: 'SpanSet'
        :return: True when this set is proper supper set of the other set.
        :rtype: bool
        """
        return other < self

    def __and__(self, other: 'SpanSet') -> 'SpanSet':
        """
        Returns sets intersection with respect to custom ∈ (in) operator of the set.

        So the definition is:
            A ∩ B = { x ∣ x∈A ∧ x∈B }   (where ∈ are custom operators for set A respective B)

        Note that the result is common set, which means that it uses exact match relation SpanSetExactEqRelation.

        :param other: The other set.
        :type other: 'SpanSet'
        :return: Intersection of both sets.
        :rtype: 'SpanSet'
        """

        return type(self)(x for x in itertools.chain(self, other) if x in self and x in other)

    def __or__(self, other: 'SpanSet') -> 'SpanSet':
        """
        Returns sets union with respect to custom ∈ (in) operator of the set.

        So the definition is:
            A ∪ B = { x ∣ x∈A ∨ x∈B }   (where ∈ are custom operators for set A respective B)

        Note that the result is common set, which means that it uses exact match relation SpanSetExactEqRelation.

        :param other:The other set.
        :type other: 'SpanSet'
        :return: Union of both sets.
        :rtype: 'SpanSet'
        """
        return type(self)(x for x in itertools.chain(self, other) if x in self or x in other)

    def __sub__(self, other) -> 'SpanSet':
        """
        Returns difference of two sets with respect to custom ∈ (in) operator of the set.

        So the definition is:
            A - B = { x ∣ x∈A ∧ x∉B }   (where ∈, ∉ are custom operators for set A respective B)

        Note that the result is common set, which means that it uses exact match relation SpanSetExactEqRelation.

        :param other:The other set.
        :type other: 'SpanSet'
        :return: Difference A - B.
        :rtype: 'SpanSet'
        """

        return type(self)(x for x in itertools.chain(self, other) if x in self and x not in other)

    def __xor__(self, other) -> 'SpanSet':
        """
        Returns symmetric difference of two sets with respect to custom ∈ (in) operator of the set.

        So the definition is:
            A ⊕ B = (A-B) ∪ (B-A) = { x| x∈A xor x∈B }(where ∈, ∉ are custom operators for set A respective B)

        Note that the result is common set, which means that it uses exact match relation SpanSetExactEqRelation.

        :param other:The other set.
        :type other: 'SpanSet'
        :return: Symmetric difference: A ⊕ B.
        :rtype: 'SpanSet'
        """

        return type(self)(x for x in itertools.chain(self, other) if (x in self) ^ (x in other))

    def isdisjoint(self, s: Iterable[Any]) -> bool:
        """
        Checks if iterable is disjoint with actual set.

        :param s: Iterable to check.
        :type s: Iterable[Any]
        :return: Returns true when all elements in iterable are not in this set.
        :rtype: bool
        """

        return all(x not in self for x in s)

    def issubset(self, other: 'SpanSet') -> bool:
        """
        Checks ifs actual set is subset of the other set.

        :param other: Set for check.
        :type other: SpanSet
        :return: True is subset. False otherwise.
        :rtype: bool
        """
        return self <= other

    def issuperset(self, other: 'SpanSet') -> bool:
        """
        Checks ifs actual set is superset of the other set.

        :param other: Set for check.
        :type other: SpanSet
        :return: True is superset. False otherwise.
        :rtype: bool
        """
        return self >= other

    def copy(self) -> 'SpanSet':
        """
        Returns shallow copy of this set.

        :return: Shallow copy.
        :rtype: 'SpanSet'
        """

        return copy.copy(self)



# -*- coding: UTF-8 -*-
""""
Created on 31.01.20
This module contains generic utils

:author:     Martin Dočekal
"""
from typing import Sequence, List, Tuple, Iterable


def get_all_subclasses(cls):
    """
    Searches all subclasses of given class.

    :param cls: The base class.
    :type cls: class
    """

    stack = [cls]
    sub = []
    while len(stack):
        base = stack.pop()
        for child in base.__subclasses__():
            if child not in sub:
                sub.append(child)
                stack.append(child)

    return sub


def sub_seq(s1: Sequence, s2: Sequence) -> bool:
    """
    Checks if sequence s1 is subsequence of s2,

    :param s1: First sequence.
    :type s1: Sequence
    :param s2: Second sequence.
    :type s2: Sequence
    :return: True if s1 is subsequence of s2.
    :rtype: bool
    """

    if len(s1) <= len(s2) and \
            any(s1 == s2[offset:offset + len(s1)] for offset in range(0, len(s2) - len(s1) + 1)):
        return True

    return False


def search_sub_seq(s1: Sequence, s2: Sequence) -> List[Tuple[int, int]]:
    """
    Searches all occurrences of sequence s1 in s2,

    :param s1: First sequence.
    :type s1: Sequence
    :param s2: Second sequence.
    :type s2: Sequence
    :return: List of searched spans. Span is a tuple [start, end).
        Empty list maybe return in case when there are no spans found.
    :rtype: List[Tuple[int, int]]
    :raise ValueError: When one of input sequences haves zero len.
    """

    if len(s1) == 0 or len(s2) == 0:
        raise ValueError("Both sequences must have non zero length.")

    if len(s1) <= len(s2):
        res = []
        for offset in range(0, len(s2) - len(s1) + 1):
            end_offset = offset + len(s1)
            if s1 == s2[offset:end_offset]:
                res.append((offset, end_offset))

        return res

    return []


class RoundSequence(object):
    """
    Wrapper for an Sequence that should iterate infinitely in cyclic fashion.
    """

    def __init__(self, i: Sequence):
        """
        Initialization of wrapper.

        :param i: Sequence you want to wrap.
        :type i: Sequence
        """

        self.s = i
        self.i = iter(self.s)

    def __iter__(self):
        return self.i

    def __next__(self, *args, **kwargs):
        try:
            x = next(self.i)
        except StopIteration:
            self.i = iter(self.s)
            x = next(self.i)

        return x


def compare_pos_in_iterables(a: Iterable, b: Iterable) -> bool:
    """
    Positionally invariant compare of two iterables.

    Example of two same iterables:
        [1,2,3]
        [3,2,1]

    Example of two different iterables:
        [1,2,3]
        [1,4,3]

    :param a: First iterable for comparison.
    :type a: Iterable
    :param b: Second iterable for comparison.
    :type b: Iterable
    :return: True considered the same. False otherwise.
    :rtype: bool
    """

    b = list(b)
    try:
        for x in a:
            b.remove(x)
    except ValueError:
        return False
    return len(b) == 0



# -*- coding: UTF-8 -*-
""""
Created on 23.09.20
This module contains classes that represents parallel workers a.k.a. processes.

:author:     Martin Dočekal
"""
import multiprocessing
from multiprocessing.context import Process
from multiprocessing import Queue
from typing import TypeVar, Callable

T = TypeVar('T')
R = TypeVar('R')


class FunRunner(Process):
    """
    Representation of one parallel process that runs given function on data from shared queue :py:attr:`~WORK_QUEUE`
    (shared among all the others FunRunners). It expects that the data is tuple
    (IS_MEANT_FOR_IDS_BUT_CAN_BY_ANYTHING_YOU_WANT, List[data]) or None.
    The None should be put into :py:attr:`~WORK_QUEUE` to terminate a FunRunner.

    The results are put into the :py:attr:`~RESULTS_QUEUE`.

    Example:
        procs = [FunRunner(pf=f) for _ in range(workers)]

        for p in procs:
            p.daemon = True
            p.start()

        dataCnt = 0

        res = []
        # push data to workers
        for i, d in enumerate(data):
            FunRunner.WORK_QUEUE.put((i, [d]))
            dataCnt += 1

            try:
                # read the results
                while True:
                    res.append(FunRunner.RESULTS_QUEUE.get(False)[0])
            except queue.Empty:
                pass

        # terminate running workers
        for _ in range(workers):
            FunRunner.WORK_QUEUE.put(None)

        # get the rest of results
        while len(res) < dataCnt:
            res.append(FunRunner.RESULTS_QUEUE.get())

        for p in procs:
            p.join()
    """

    WORK_QUEUE = Queue(multiprocessing.cpu_count())
    """Queue for work that needs to be done."""

    RESULTS_QUEUE = Queue()
    """Completed work."""

    def __init__(self, pf: Callable[[T], R]):
        """
        Initialization of parallel worker.

        :param pf: Function you want to run in data-parallel way.
        :type pf: Callable[[T], R]
        """
        super().__init__()
        self.pf = pf

    def run(self) -> None:
        """
        Run the process.
        """
        try:
            while True:
                q_item = self.WORK_QUEUE.get()

                if q_item is None:
                    # all done
                    break

                i, data_list = q_item

                self.RESULTS_QUEUE.put((i, [self.pf(x) for x in data_list]))

        finally:
            self.WORK_QUEUE.close()
            self.RESULTS_QUEUE.close()



# -*- coding: UTF-8 -*-
""""
Created on 30.06.20

Module containing class that stores configuration.

:author:     Martin Dočekal
"""
import ast
import os
from pathlib import Path
from typing import Dict


class Config(dict):
    """
    Base config class without validation of input parameters.
    If you want to validate the input override the validate method.

    Loads configuration from the python file containing only single dictionary.
    This dictionary is safely loaded with the ast.literal_eval (https://docs.python.org/3/library/ast.html).
    """

    def __init__(self, path_to: str):
        """
        Config initialization.

        :param path_to: Path to .py file with configuration.
        :type path_to: str
        :raise SyntaxError: Invalid input.
        :raise ValueError: Invalid value for a parameter or missing parameter.
        """
        self._path_to = Path(path_to).absolute()
        with open(path_to, "r") as f:
            config = ast.literal_eval(f.read())

            if not isinstance(config, dict):
                raise SyntaxError("The configuration must be dict.")

            self.validate(config)

            super().__init__(config)

    def translate_file_path(self, path: str) -> str:
        """
        Translates relative path to the absolute path.
        If the path is already absolute than it returns the original.
        All relative paths are set relatively to this config file.

        :param path: Path for translating.
        :type path: str
        :return: translated path
        :rtype: str
        """
        if not os.path.isabs(path):
            path = os.path.join(os.path.dirname(self._path_to), path)

        return path

    def validate(self, config: Dict):
        """
        Validates the loaded configuration.

        :param config: Loaded configuration.
        :type config: Dict
        :raise ValueError: Invalid value for a parameter or missing parameter.
        """

        pass



# -*- coding: UTF-8 -*-
""""
Created on 26.01.21
Module containing map structures.

:author:     Martin Dočekal
"""
import bisect
from typing import Dict, Tuple, Any, Generator

from windpyutils.structures.span_set import SpanSetOverlapsEqRelation, SpanSet


class ImmutIntervalMap:
    """
    Immutable mapping of disjunctive intervals to values.

    Beware that getting value from ends of an interval could be problematic due to float numbers.

    Usage:
        iMap = IntervalMap({(1,100): "first interval", (200,201.5): "second interval"})
        iMap[5]
        "first interval"
    """

    def __init__(self, mapping: Dict[Tuple[float, float], Any]):
        """
        Initialization of a mapping.

        :param mapping: Dictionary that defines the interval mapping.
        Keys must be disjunctive intervals that are defined with a start and end of interval (both inclusive).
        :type mapping: Dict[Tuple[float, float], Any]
        :raise KeyError: When key intervals are not disjunctive or are invalid.
        """

        self._interval_starts = []
        interval_ends = []
        self._values = []
        for (start, end), value in mapping.items():
            if start > end:
                raise KeyError(f"The interval [{start}, {end}] is invalid.")

            self._interval_starts.append(start)
            interval_ends.append(end)
            self._values.append(value)

        # check that intervals are disjunctive
        span_set = SpanSet(self._interval_starts, interval_ends, eq_relation=SpanSetOverlapsEqRelation())

        if len(span_set) != len(mapping):
            raise KeyError("The key intervals are not disjunctive.")

        # this structure is main part used in searching
        self._sortedEnds = []
        self._sortedEndsIndices = []
        for i, e in sorted(enumerate(interval_ends), key=lambda x: x[1]):
            self._sortedEndsIndices.append(i)
            self._sortedEnds.append(e)

    def __len__(self):
        return len(self._interval_starts)

    def __getitem__(self, key: float) -> Any:
        """
        Searches in which interval a given key is.

        :param key: A value from an stored interval.
        :type key: float
        :return: Value associated to the interval in which the given key is.
        :rtype: Any
        """
        # search smallest interval ends that is grater or equal to key
        searched_i = bisect.bisect_left(self._sortedEnds, key)

        if searched_i == len(self._sortedEnds):
            raise KeyError("Provided key is not in any interval.")

        interval_index = self._sortedEndsIndices[searched_i]
        interval_start = self._interval_starts[interval_index]

        if key < interval_start:
            raise KeyError("Provided key is not in any interval.")

        return self._values[interval_index]

    def __iter__(self) -> Generator[Tuple[Tuple[float, float], Any], None, None]:
        """
        Iterates over key intervals and their values. The keys are iterated in ascending order.

        :return: The generator retuns:
                ((interval start, interval end), associated value with interval)
        :rtype: Generator[Tuple[Tuple[float, float], Any], None, None]
        """

        for i, e in zip(self._sortedEndsIndices, self._sortedEnds):
            yield (self._interval_starts[i], e), self._values[i]

    def __contains__(self, key: float):
        try:
            _ = self[key]
            return True
        except KeyError:
            return False



# -*- coding: UTF-8 -*-
""""
Created on 23.03.20
Circular buffer implementation.

:author:     Martin Dočekal
"""
from typing import Any, Sequence


class CircularBuffer(Sequence):
    """
    Implementation of circular buffer.
    """

    def __init__(self, max_size: int):
        """
        Implementation of circular butter.

        :param max_size: The max size of circular buffer.
            Must be grater than 0.
        :type max_size: int
        :raise AssertionError: When you provide invalid max size.
        """

        assert max_size > 0

        self._buffer = [None] * max_size
        self._size = 0
        self._offset = 0  # Offset for next element that will be added.

    def __len__(self) -> int:
        """
        Actual size of buffer.

        :return: Actual size of buffer.
        :rtype: int
        """

        return self._size

    def __getitem__(self, offset: int) -> Any:
        """
        Get item on offset.

        Time complexity: O(1)

        :param offset: The offset of item you want.
        :type offset: int
        :return: Element on given offset.
        :rtype: Any
        :raise IndexError: when you touch outside the actual size of buffer.
        """

        if offset >= len(self) or 0 > offset:
            raise IndexError("The offset {} is out of buffer.".format(offset))

        return self._buffer[(self._offset-self._size+offset) % self.max_size]

    @property
    def max_size(self) -> int:
        """
        Maximum buffer size
        """

        return len(self._buffer)

    def put(self, e: Any):
        """
        Put element into circular buffer.

        Time complexity: O(1)

        :param e: Element you want to add.
        :type e: Any
        """

        self._buffer[self._offset] = e
        self._offset = (self._offset + 1) % self.max_size

        if self._size < self.max_size:
            self._size += 1

    def clear(self):
        """
        Removes all items from buffer.

        Time complexity: O(1)
        """

        self._size = 0
        self._offset = 0



# -*- coding: UTF-8 -*-
""""
Created on 05.02.21
Visualization tools for text environment.

:author:     Martin Dočekal
"""
import math
import sys
from typing import Dict, TextIO, Tuple, Sequence, Optional


def print_histogram(bars: Sequence[Tuple[str, float]], max_width: int = 40, file: Optional[TextIO] = None,
                    bar_char: str = "█", print_value: bool = True):
    """
    Prints histogram defined by the provided bars.

    Example:
        >>> print_histogram([("red", 10), ("green", 5), ("blue", 7)])
        red   ████████████████████████████████████████ 10
        green ████████████████████ 5
        blue  ████████████████████████████ 7

    :param bars: It's sequence of bars. Bar is a tuple. Tuple defines label and value associated with it.
        Value must be always non negative.
    :type bars: Sequence[Tuple[str, float]]
    :param max_width: Maximal number of bar_chars that represents histogram bar.
    :type max_width: int
    :param file: Prints results to this output.
    :type file: Optional[TextIO]
    :param bar_char: Character (or multiple) that will be used for creating representation of a bar.
    :type bar_char: str
    :param print_value: If true prints value after the bar.
    :type print_value: True
    :raise AssertionError: On invalid max_width.
    :raise ValueError: When a value is negative.
    """
    if file is None:
        file = sys.stdout

    max_label_chars = 0
    max_value = - math.inf

    for label, value in bars:
        if len(label) > max_label_chars:
            max_label_chars = len(label)
        if value < 0:
            raise ValueError(f"The value {value} for label {label} is negative.")
        if value > max_value:
            max_value = value

    for label, value in bars:
        formatting_spaces = " " * (max_label_chars - len(label))
        bar = bar_char * (round(max_width * value / max_value))

        print(f"{label}{formatting_spaces} {bar}{' ' + str(value) if print_value else ''}", file=file)


def print_buckets_histogram(values: Dict[float, float], buckets: int = -1, bucket_size_int: bool = False,
                            max_width: int = 40, file: Optional[TextIO] = None, bar_char: str = "█",
                            print_value: bool = True, decimals: int = 2):
    """
    Prints histogram defined by the provided dictionary the x values are the keys and y values are the dictionary values
    for given key. The keys will be automatically sorted into buckets if buckets > 0. Each bucket is associated with one
    histogram bar.

    The bar label will the associated key in case when buckets == -1, else it will be the buckets interval.

    Example:
        >>> print_buckets_histogram({0:2,5:8, 10:18, 15:8, 20:2}, buckets=1)
        [0,20] ████████████████████████████████████████ 38

        >>> print_buckets_histogram({0:2,5:8, 10:18, 15:8, 20:2}, buckets=3)
        [0,6.67)     ██████████████████████ 10
        [6.67,13.33) ████████████████████████████████████████ 18
        [13.33,20]   ██████████████████████ 10

        >>> print_buckets_histogram({0:2,5:8, 10:18, 15:8, 20:2}, buckets=5)
        [0,4)   ████ 2
        [4,8)   ██████████████████ 8
        [8,12)  ████████████████████████████████████████ 18
        [12,16) ██████████████████ 8
        [16,20] ████ 2

        >>> print_buckets_histogram({0:2,5:8, 10:18, 15:8, 20:2})
        0  ████ 2
        5  ██████████████████ 8
        10 ████████████████████████████████████████ 18
        15 ██████████████████ 8
        20 ████ 2

    :param values: Dictionary that defines the x,y coordinates for the histogram.
    :type values: Dict[float, float]
    :param buckets: Number of histogram buckets. Its maximal number of buckets if there would be a single key value
    then only one bucket will be created even though buckets would be equal to 10.
        -1 means that number of buckets should be equal to number of keys in values dict
    :type buckets: int
    :param bucket_size_int: If true bucket size will be rounded up to nearest integer (just for buckets > 0).
        May cause that the last bucket interval will end after max.
    :type bucket_size_int: bool
    :param max_width: Maximal number of characters that represents histogram bar.
    :type max_width: int
    :param file: Prints results to this output.
    :type file: Optional[TextIO]
    :param bar_char: Character (or multiple) that will be used for creating representation of a bar.
    :type bar_char: str
    :param print_value: If true prints value after the bar.
    :type print_value: True
    :param decimals: Float numbers will be rounded to x decimal points.
    :type decimals: int
    :raise AssertionError: On invalid number of buckets, on invalid max_width or empty values dict.
    :raise ValueError: When a value in bucket is negative.
    """

    assert buckets == -1 or buckets > 0
    assert len(values) > 0

    if buckets == -1 or len(values) == 1:
        # buckets are already defined
        print_histogram([(str(label), val) for label, val in sorted(values.items(), key=lambda ite: ite[0])],
                        max_width, file, bar_char, print_value)
    else:
        # we need to sort keys into buckets
        min_k = min(values.keys())
        max_k = max(values.keys())
        keys_spread = max_k - min_k
        bucket_size = keys_spread / buckets
        if bucket_size_int:
            bucket_size = math.ceil(bucket_size)

        hist = [0] * buckets

        for k, v in values.items():
            belongs_to_bucket = int((k - min_k) / bucket_size)
            if belongs_to_bucket == buckets:  # only when the k is max
                # the last bucket interval is closed and we have max value so it belongs to the las bucket
                belongs_to_bucket -= 1

            hist[belongs_to_bucket] += v

        bucket_labels = []
        bucket_start = min_k
        for bucket_offset in range(len(hist) - 1):  # the last one is handled separately
            bucket_end = bucket_start + bucket_size
            str_s = f"{bucket_start:.{decimals}f}".rstrip("0").rstrip(".")
            str_e = f"{bucket_end:.{decimals}f}".rstrip("0").rstrip(".")
            bucket_labels.append(f"[{str_s},{str_e})")
            bucket_start = bucket_end

        # the last one is closed interval to the max value
        bucket_end = max_k
        if bucket_size_int:
            bucket_end = math.ceil(bucket_end)

        str_s = f"{bucket_start:.{decimals}f}".rstrip("0").rstrip(".")
        str_e = f"{bucket_end:.{decimals}f}".rstrip("0").rstrip(".")
        bucket_labels.append(f"[{str_s},{str_e}]")

        print_histogram([(label, val) for label, val in zip(bucket_labels, hist)],
                        max_width, file, bar_char, print_value)



# -*- coding: UTF-8 -*-
""""
Created on 23.09.20
Module with mapping functions.

:author:     Martin Dočekal
"""
import multiprocessing
import queue
from typing import TypeVar, Callable, Iterable, List

from windpyutils.parallel.workers import FunRunner

T = TypeVar('T')
R = TypeVar('R')


def mul_p_map(f: Callable[[T], R], data: Iterable[T], workers: int = -1) -> List[R]:
    """
    Runs function f with arguments X

    :param f: Function you want to run in data-parallel way
    :type f: Callable[[T], R]
    :param data: The data that will be processed.
    :type data:Iterable[T]
    :param workers: Number of parallel workers.
        Values <=0 will create number of workers that will be same as number of cpus.
    :type workers: int
    :return: Processed input.
    :rtype:
    """
    if workers <= 0:
        workers = multiprocessing.cpu_count()

    procs = [FunRunner(pf=f) for _ in range(workers)]

    for p in procs:
        p.daemon = True
        p.start()

    data_cnt = 0

    res = []

    for i, d in enumerate(data):
        FunRunner.WORK_QUEUE.put((i, [d]))
        data_cnt += 1

        try:
            # read the results
            while True:
                act = FunRunner.RESULTS_QUEUE.get(False)
                res.append((act[0], act[1][0]))
        except queue.Empty:
            pass

    for _ in range(workers):
        FunRunner.WORK_QUEUE.put(None)

    while len(res) < data_cnt:
        act = FunRunner.RESULTS_QUEUE.get()
        res.append((act[0], act[1][0]))

    for p in procs:
        p.join()

    return [r for i, r in sorted(res, key=lambda x: x[0])]



# -*- coding: UTF-8 -*-
""""
Created on 23.12.19

:author:     Martin Dočekal
"""
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

setup_args = dict(
    name='windpyutils',
    version='1.2.2',
    description='Useful tools for Python projects.',
    long_description_content_type="text/markdown",
    long_description=README,
    license='The Unlicense',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    author='Martin Dočekal',
    keywords=['utils', 'general usage'],
    url='https://github.com/mdocekal/windPyUtils',
    python_requires='>=3.6',
    install_requires=[]
)

if __name__ == '__main__':
    setup(**setup_args)


# -*- coding: UTF-8 -*-
""""
Created on 13.02.20
This module contains metrics.

:author:     Martin Dočekal
"""
import math
from typing import Iterable


def mean_squared_error(results: Iterable[float], targets: Iterable[float]) -> float:
    """
    Calculates mean squered error

    :param results: Guessed results.
    :type results: Iterable[float]
    :param targets: Ground truth results.
    :type targets: Iterable[float]
    :return: MSE
    :rtype: float
    """

    sum_val = 0
    cnt = 0

    for r, t in zip(results, targets):
        sum_val += (r-t)**2
        cnt += 1

    return sum_val/cnt


def root_mean_squared_error(results: Iterable[float], targets: Iterable[float]) -> float:
    """
    Calculates root mean squered error

    :param results: Guessed results.
    :type results: Iterable[float]
    :param targets: Ground truth results.
    :type targets: Iterable[float]
    :return: RMSE
    :rtype: float
    """

    return math.sqrt(mean_squared_error(results, targets))
